package net.knarcraft.placeholdersigns.listener;

import net.knarcraft.knarlib.formatting.StringFormatter;
import net.knarcraft.knarlib.formatting.StringReplacer;
import net.knarcraft.knarlib.property.ColorConversion;
import net.knarcraft.knarlib.util.ColorHelper;
import net.knarcraft.placeholdersigns.PlaceholderSigns;
import net.knarcraft.placeholdersigns.config.PlaceholderSignMessage;
import net.knarcraft.placeholdersigns.container.PlaceholderSign;
import net.knarcraft.placeholdersigns.container.SignLineChangeRequest;
import net.knarcraft.placeholdersigns.handler.PlaceholderSignHandler;
import net.knarcraft.placeholdersigns.handler.PlaceholderSignRequestHandler;
import net.md_5.bungee.api.ChatColor;
import org.bukkit.Bukkit;
import org.bukkit.DyeColor;
import org.bukkit.Location;
import org.bukkit.block.Sign;
import org.bukkit.block.sign.Side;
import org.bukkit.block.sign.SignSide;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.SignChangeEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.jetbrains.annotations.NotNull;

import java.util.Map;
import java.util.Objects;

/**
 * A listener for placeholder signs being clicked
 */
public class SignClickListener implements Listener {

    @EventHandler(priority = EventPriority.MONITOR)
    public void onSignClick(@NotNull PlayerInteractEvent event) {
        // Ignore if not a clicked sign
        if (!event.hasBlock() || event.getClickedBlock() == null ||
                !(event.getClickedBlock().getState() instanceof Sign sign)) {
            return;
        }

        Player player = event.getPlayer();
        PlaceholderSignRequestHandler requestHandler = PlaceholderSigns.getInstance().getRequestHandler();

        Boolean hasSignViewRequest = requestHandler.getSignViewRequest(player);
        if (hasSignViewRequest != null) {
            printSign(sign, player, hasSignViewRequest, false);
            // Cancel the event to prevent vanilla behavior
            event.setCancelled(true);
            return;
        }

        Boolean hasPlaceholderSignViewRequest = requestHandler.getPlaceholderSignViewRequest(player);
        if (hasPlaceholderSignViewRequest != null) {
            printSign(sign, player, hasPlaceholderSignViewRequest, true);
            event.setCancelled(true);
            return;
        }

        // Check if the player has run the /editSign command
        SignLineChangeRequest request = requestHandler.getSignChangeRequest(player);
        if (request != null) {
            if (sign.isWaxed() && !player.hasPermission("placeholdersigns.edit.bypass-waxed")) {
                PlaceholderSigns.getInstance().getStringFormatter().displayErrorMessage(player,
                        PlaceholderSignMessage.ERROR_WAXED_NO_PERMISSION);
                return;
            }

            SignSide standingOn = sign.getTargetSide(player);
            Side side = sign.getSide(Side.FRONT).equals(standingOn) ? Side.FRONT : Side.BACK;
            doSignChange(sign, request, event, player, side);
        }
    }

    /**
     * Prints the current contents of a sign to a player
     *
     * @param sign             <p>The sign to print</p>
     * @param player           <p>The player to display the contents to</p>
     * @param raw              <p>Whether to get the raw text with used formatting codes</p>
     * @param showPlaceholders <p>Whether to show the actual placeholders stored on the sign</p>
     */
    private void printSign(@NotNull Sign sign, @NotNull Player player, boolean raw, boolean showPlaceholders) {
        Location location = sign.getLocation();
        SignSide front = sign.getSide(Side.FRONT);
        SignSide back = sign.getSide(Side.BACK);
        String frontLines = showPlaceholders ? getPlaceholderSignText(location, front, Side.FRONT, raw) :
                getSignText(front.getLines(), raw);
        String backLines = showPlaceholders ? getPlaceholderSignText(location, back, Side.BACK, raw) :
                getSignText(back.getLines(), raw);

        StringFormatter stringFormatter = PlaceholderSigns.getInstance().getStringFormatter();
        StringReplacer replacer = new StringReplacer(stringFormatter.getUnformattedColoredMessage(
                PlaceholderSignMessage.SUCCESS_SIGN_CONTENTS));
        replacer.add("{frontLines}", frontLines);
        replacer.add("{backLines}", backLines);
        replacer.add("{frontDye}", getDye(front));
        replacer.add("{frontGlow}", getStatus(front.isGlowingText()));
        replacer.add("{backDye}", getDye(back));
        replacer.add("{backGlow}", getStatus(back.isGlowingText()));
        replacer.add("{waxed}", getStatus(sign.isWaxed()));
        player.sendMessage(replacer.replace());
    }

    /**
     * Gets a description of the status of a sign property
     *
     * @param isTrue <p>Whether the property is true</p>
     * @return <p>A description of the property's current status</p>
     */
    @NotNull
    private String getStatus(boolean isTrue) {
        StringFormatter stringFormatter = PlaceholderSigns.getInstance().getStringFormatter();
        if (isTrue) {
            return stringFormatter.getUnformattedColoredMessage(PlaceholderSignMessage.SIGN_PROPERTY_CONFIRM);
        } else {
            return stringFormatter.getUnformattedColoredMessage(PlaceholderSignMessage.SIGN_PROPERTY_DENY);
        }
    }

    /**
     * Gets a description of a dye applied to a sign
     *
     * @param signSide <p>The sign side to get the dye of</p>
     * @return <p>The description of the applied dye</p>
     */
    @NotNull
    private String getDye(@NotNull SignSide signSide) {
        DyeColor dyeColor = signSide.getColor();
        if (dyeColor == null) {
            return "None";
        }

        ChatColor color = ColorHelper.fromColor(dyeColor.getColor());
        return color + signSide.getColor().name();
    }

    /**
     * Gets text from a sign side, and appends it to the given string builder
     *
     * @param signLocation <p>The location of the sign</p>
     * @param signSide     <p>The sign side to get text from</p>
     * @param side         <p>The side of the sign to get placeholders from</p>
     * @param raw          <p>Whether to get the raw text with used formatting codes</p>
     */
    @NotNull
    private String getPlaceholderSignText(@NotNull Location signLocation, @NotNull SignSide signSide,
                                          @NotNull Side side, boolean raw) {
        String[] lines = signSide.getLines();
        PlaceholderSign placeholderSign = PlaceholderSigns.getInstance().getSignHandler().getFromLocation(signLocation);
        if (placeholderSign == null) {
            return getSignText(lines, raw);
        }

        Map<Integer, String> placeholders = placeholderSign.placeholders().get(side);
        for (Map.Entry<Integer, String> entry : placeholders.entrySet()) {
            lines[entry.getKey()] = entry.getValue();
        }

        return getSignText(lines, raw);
    }

    /**
     * Gets text from a sign side, and appends it to the given string builder
     *
     * @param lines <p>The lines on the sign</p>
     * @param raw   <p>Whether to get the raw text with used formatting codes</p>
     */
    @NotNull
    private String getSignText(@NotNull String[] lines, boolean raw) {
        StringBuilder output = new StringBuilder();
        for (int i = 0; i < 4; i++) {
            output.append(i + 1).append(". ");

            String line = lines[i];
            if (raw) {
                output.append(line.replace(ChatColor.COLOR_CHAR, '&'));
            } else {
                output.append(line);
            }
            output.append(ChatColor.COLOR_CHAR).append("r\n");
        }
        return output.toString();
    }

    /**
     * Perform the changing of a sign according to the given sign change request
     *
     * @param sign    <p>The sign to be changed</p>
     * @param request <p>The sign line change request to perform</p>
     * @param event   <p>The interaction event that triggered this</p>
     * @param player  <p>The player triggering the sign change</p>
     */
    private void doSignChange(@NotNull Sign sign, @NotNull SignLineChangeRequest request,
                              @NotNull PlayerInteractEvent event, @NotNull Player player, @NotNull Side side) {
        // Cancel the event to prevent vanilla behavior
        event.setCancelled(true);

        SignSide signSide = sign.getSide(side);
        String[] lines = signSide.getLines();

        PlaceholderSignHandler signHandler = PlaceholderSigns.getInstance().getSignHandler();
        PlaceholderSign placeholderSign = PlaceholderSigns.getInstance().getSignHandler().getFromLocation(sign.getLocation());
        String oldPlaceholder = null;
        if (placeholderSign != null) {
            // Remove the old placeholder
            oldPlaceholder = placeholderSign.placeholders().get(side).remove(request.line());
        }

        lines[request.line()] = request.text();

        // Run the sign change event to allow protection plugins to cancel, and allow the sign text listener to trigger
        SignChangeEvent changeEvent = new SignChangeEvent(Objects.requireNonNull(event.getClickedBlock()),
                player, lines, side);
        Bukkit.getPluginManager().callEvent(changeEvent);
        if (changeEvent.isCancelled()) {
            if (placeholderSign != null) {
                // Restore the old placeholder if the action didn't complete
                placeholderSign.placeholders().get(side).put(request.line(), oldPlaceholder);
            }
            return;
        }

        // Update the sign with the new text
        String[] finalLines = changeEvent.getLines();
        for (int i = 0; i < finalLines.length; i++) {
            signSide.setLine(i, ColorHelper.translateColorCodes(finalLines[i], ColorConversion.RGB));
        }
        sign.update();

        // Save any placeholder changes
        signHandler.save();

        PlaceholderSigns.getInstance().getStringFormatter().displaySuccessMessage(player,
                PlaceholderSignMessage.SUCCESS_SIGN_CHANGED);
    }

}
